# Internationalization and Localization

## Translations

### Terminologies of translations

In the following, you find some details about the terminology of translations.

- **catalog** is a key/value store for translations.
- **key** is a unique identifier used to lookup code translations in the catalog.
- **locale** is a two letter language code (for example `de` or `fr`).
- **sentence** is a translation for a given key. A single key might have multiple translations .
- **scope** is a translation domain grouping the translation data related to a skill or service. The **scope** is usually equals to the the service/skill name.
- **.po/.mo files** - translation files in gettext format.

### Conventions for translations

Use two letter [ISO 639-1 codes](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes) for the language of translations.

Do not use any country or ISO 3166-1 codes:

- :no_entry_sign: `de_DE`
- :no_entry_sign: `deu`
- :no_entry_sign: `ger`
- :white_check_mark: `de`

### Loading of translations

#### Local gettext files

Local gettext translation files are expected in the `locale/` folder. For example, a skill that supports English, French and German language has the following directory structure:

```
skill-root/
└── locale/
    ├── de.po
    ├── en.po
    └── fr.po
```

#### Text service

A central text service provides the translation data. An API wrapper is implemented in `skill_sdk.services.text.TextService`.

The text service is configured in `[service-text]` section of `skill.conf` file:

```ini
[service-text]
url = http://service-text-service:8080
active = [true | false]             (true)
auth_header = [false | true]        (false)
load = [auto | delayed | startup]   (auto)
```

The text services endpoint URL is set by `url` parameter. Default value for locally deployed skill is `http://service-text-service:8080`.

To explicitly deactivate the service, set the `active` parameter to `false`. Default value is `true. So if you have just 
added the `[service-text]` section to your skill configuration file, the service is activated during the skill deployment.

If the `auth_header` value is `true`, an additional HTTP header `Authorization: Bearer {cvi_token}` is added to every request to the text services. 
This header is necessary if skill running outside of DTAG cloud wants to access the text services via device gateway.

#### Process of translation loading

When skill is being deployed, the translations are loaded in the following order:

1. Scan the projects subfolder `locale/` for `.po` files.
2. Compile `.po` files to `.mo` file.
3. Load the `.mo` files as translations.
4. Request the list of supported locales for this scope from the text service.
5. Load the translations from the text service and replace the `gettext` translations.

The `load` parameter in `[service-text]` section defines **when** the translations are requested from text services:

    "auto":     the translations are loaded at skill startup and reloaded at TEXT_SERVICE_UPDATE_INTERVAL
                a skill will start without translations if text services are unavailable
                (this is a default behaviour)

    "delayed":  the skill will start without translations and loads the catalog when requested by CVI
                (i.e on a first invocation, if the translations are empty)

    "startup":  the translations are loaded at skill startup and reloaded at TEXT_SERVICE_UPDATE_INTERVAL
                a skill will raise a l10n.TranslationError exception if text services are not available

### Usage in code

The translation that matches the locale of the request is injected into the `context` as `translation`.

The translation supports the following methods:

- `gettext()`
- `ngettext()`
- `getalltexts()` (this method returns a list of all available texts for a specific tag and language)

>You find a description of these methods in the [gettext.GNUTranslations API docs](https://docs.python.org/3/library/gettext.html#the-gnutranslations-class).

For convenience purposes, the following shortcuts exist:

- `context._()` for `context.translation.gettext()`
- `context._n()` for `context.translation.ngettext()`
- `context._a()` for `context.translation.getalltexts()`

Global `l10n._`/`l10n._n`/`l10n._a` functions are available. They are thread-safe and if accessed within a request invocation context, 
these functions always refers to the currently requested locale.
 
#### Examples

```python
from skill_sdk.l10n import _
from skill_sdk.skill import intent_handler

@intent_handler('HELLO_INTENT')
def hello():
    return _('HELLO')
```

```python
from skill_sdk.l10n import _, _n, _a

card_title = _('WEATHER_CARD_TITLE')
text = _n('WEATHER_TEXT_ONE_DAY', 'WEATHER_TEXT_MULTIPLE_DAYS', num_days)
all_texts = _a('THANKS')
```
You can create a response message with format arguments by using a constructor or `format` function. The following two statements are equivalent:
```python
text = _('WEATHER_TEXT', temperature=temp, forecast=forecast)
text = _('WEATHER_TEXT').format(temperature=temp, forecast=forecast)
```

### Usage in tests

When you write tests or test your app with Python unit test modules, consider that during unit testing the text service does not load the actual translation. 
In this case, the translation system uses your local `gettext` translation in the `locale` folder or, if no `locale` folder is available, no translation at all.

**Example**

If you create your message by using the `format` function `text = context._('WEATHER_TEXT').format(temperature=temp, forecast=forecast)`,
the `text` variable contains additional fields:
```python
>>> type(text)
<class 'skill_sdk.l10n.Message'>
>>> text.key
'WEATHER_TEXT'
>>> text.kwargs
{'temperature': temp, 'forecast': forecast}
```

To avoid issues when a multi-locale translation is available, use the following fields in your unit test: 

```python
self.assertEqual(text.key, 'WEATHER_TEXT')
```

Alternatively, you can nullify the translations before executing the unit tests:
```python
from skill_sdk.test_helpers import set_translations
set_translations(None)   
```

This function call sets all available translations to `gettext.NullTranslations` object. 
 
## Templating

To format in Python, the sentences can use the `.format()` string.
For a proper documentation, click [here](https://pyformat.info/).

#### *New in version 1.2.4*

You can now extract translatable strings from Python files or modules with `python[3] manage.py translate [modules]`
command. As a result it creates a `locale/messages.pot` translation template in .POT format.
